/*
* bb_util.c - utility procedures for blockbuilder.
*/

#include "bb_util.h"

/* ------------------ block-printing for debugging -------------------------- */

void show_block(Block_ptr b, FILE *debug_file)
{
	int i, j, J, c = b->ncol, r = b->nrow, W = 20, beg, end;
	Seg_ptr s = b->seg;
	uchar x, *t;
	char buf[16];

	fprintf(debug_file, " block of %d rows, %d columns\n", r, c);
	for (i = 0; i < r; ++i) {
		beg = s[i].beg;
		end = s[i].end;
		snprintf(buf, 15, "%d-%d", beg, end);
		fprintf(debug_file, "  contig %d (+) %15s  ",
		   s[i].contig_nbr, buf);
		t = s[i].chars;
		if (c <= 50) {
			for (j = 0; j < c; ++j) {
				x = ((i > 0 && t[j] == s[0].chars[j] &&
				   t[j] != '-') ? '.' : t[j]);
				putc(x, debug_file);
			}
			putc('\n', debug_file);
		} else {
			for (j = 0; j < W; ++j) {
				x = ((i > 0 && t[j] == s[0].chars[j] &&
				   t[j] != '-') ? '.' : t[j]);
				putc(x, debug_file);
			}
			fprintf(debug_file, " .. ");
			for (j = 0; j < W; ++j) {
				J = c - W + j;
				x = ((i > 0 && t[J] == s[0].chars[J] &&
				   t[J] != '-') ? '.' : t[J]);
				putc(x, debug_file);
			}
			putc('\n', debug_file);
		}
	}
}

void show_blocks(Block_ptr b, FILE *debug_file)
{
	for (; b; b = b->next)
		show_block(b, debug_file);
}

/* ------------------------- utility routines ------------------------------- */

int col_of_pos(int pos, Seg_ptr s, int slen)
{
	int col, p;
	uchar *t;

	if (pos < s->beg || s->end < pos)
		fatal("col_of_pos: improper pos");
	p = s->beg - 1;
	for (col = 0, t = s->chars; p < pos && col < slen; ++col, ++t)
		if (*t != '-' && ++p >= pos)
			break;
	if (col == slen)
		fatal("col_of_pos: cannot find pos");
	return col;
}

/* pos_of_col returns the first sequence position at or after a given column
*  of a given segment */
int pos_of_col(int col, Seg_ptr s, int slen)
{
	int pos = s->end + 1;
	uchar *t = s->chars + slen;

	while (--slen >= col)
		if (*--t != '-')
			--pos;
	return pos;
	
}

Seg_ptr find_seg(int contig_nbr, int pos, Contig_ptr contig, Block_ptr b,
  int must)
{
	int i, len;
	Seg_ptr s;

	if (b == NULL)
		fatal("find_seg: b is NULL");
	len = contig[contig_nbr].len + 1;
	for (i = 0; i < b->nrow; ++i) {
		s = &(b->seg[i]);
		if (s->contig_nbr == contig_nbr &&
		  s->beg <= pos && s->end >= pos)
			return s;
	}
	if (must)
		fatal("find_seg: cannot find sequence in block");
	return NULL;
}

/* Find a block containing a specified position in a specified orientation of a
*  specified contig.  Reverse the block if necessary.
*/
Block_ptr find_block(Block_ptr b, int contig_nbr, int pos, Contig_ptr contig,
  Seg_ptr *S)
{
	Block_ptr found = NULL;
	Seg_ptr s;

	for ( ; b; b = b->next)
		if ((s = find_seg(contig_nbr, pos, contig, b, 0))
		     != NULL)
			if (s->beg <= pos && s->end >= pos) {
				if (found != NULL)
				  fatalf("contig `%s' pos %d in two blocks",
					contig[contig_nbr].header, pos);
				else {
					found = b;
					*S = s;
				}
			}
	if (found == NULL)
		fatalf("find_block: cannot find contig %d, pos %d\n",
			contig_nbr, pos);

	s = *S;
	b = found;

	return b;
}

Block_ptr adjacent_block(Seg_ptr *S, Contig_ptr contig, Block_ptr blocks, 
  int forward)
{
    	if (forward)
                return (next_block(S, contig, blocks));
	else
		return (prev_block(S, contig, blocks));
}

Block_ptr prev_block(Seg_ptr *S, Contig_ptr contig, Block_ptr blocks)
{
        Seg_ptr s = *S;
        Block_ptr b;
        int contig_nbr, len, next_pos;

        contig_nbr = s->contig_nbr;
        len = contig[contig_nbr].len;
        next_pos = s->beg - 1;
        if (next_pos > 0)
                b = find_block(blocks, contig_nbr, next_pos, contig, S);
        else {
                b = NULL;
                *S = NULL;
        }
        return b;
}

Block_ptr next_block(Seg_ptr *S, Contig_ptr contig, Block_ptr blocks)
{
	Seg_ptr s = *S;
	Block_ptr b;
	int contig_nbr, len, next_pos;

	contig_nbr = s->contig_nbr;
	len = contig[contig_nbr].len;
	next_pos = 1 + s->end;
	if (next_pos <= len)
		b = find_block(blocks, contig_nbr, next_pos, contig, S);
	else {
		b = NULL;
		*S = NULL;
	}
	return b;
}

/* write_blocks -- write blocks, possibly in compresed form to stdout.
*  Check that *  the number of non-dash characters in each row agrees with the
*  declared beg and end positions.
*/
void write_blocks(Block_ptr b, Seq_ptr seq, int nseq, Contig_ptr contig,
  int ncontig, int verbose)
{
	int i, j, k, total;
	Seg s;
	char *t;

	printf("%d %d\n", nseq, ncontig);
	for (i = 0; i < ncontig; ++i)
		printf("%s \"%s\" %d\n", seq[contig[i].seq_nbr].name,
			contig[i].header, contig[i].len);

	for ( ; b; b = b->next) {
		putchar('\n');
		printf("%d %d\n", b->nrow, b->ncol);
		for (i = 0; i < b->nrow; ++i) {
			s = b->seg[i];
			printf("%d ", s.contig_nbr);
			printf("+ %d-%d", s.beg, s.end);
			if (verbose) {
				putchar(' ');
				for (j = 0; j < b->ncol; ++j)
					putchar(s.chars[j]);
				putchar('\n');
				continue;
			}
			for (total = j = k = 0, t = s.chars; j < b->ncol;
			  ++j, ++t)
				if (*t == '-') {
					if (k > 0) {
						printf(" %d", k);
						total += k;
						k = 0;
					}
					--k;
				} else {
					if (k < 0) {
						printf(" %d", k);
						k = 0;
					}
					++k;
				}
			printf(" %d\n", k);
			if (k > 0)
				total += k;
			if (s.end - s.beg + 1 != total) {
				fprintf(stderr, "%d - %d + 1 != %d; ",
				  s.end, s.beg, total);
				fatal("see last line of output");
			}
		}
	}

}

static void read_seqs(Seq_ptr seq, int nseq, Contig_ptr contig, int ncontig)
{
	SEQ *sf;
	int i, j;

	for (j = i = 0; i < nseq; ++i) {
		sf = seq_open(seq[i].name);
		while (seq_read(sf)) {
			if (contig[j].seq_nbr != i)
				fatalf("Disagreeing contigs in %s",
				  seq[i].name);
			if (!same_string(contig[j].header,SEQ_HEAD(sf)))
				fatal("mismatched contig headers");
			if (contig[j].len != SEQ_LEN(sf))
				fatal("mismatched contig lengths");
			contig[j].chars = copy_string(SEQ_CHARS(sf));
			++j;
		}
		seq_close(sf);
	}
	if (j != ncontig)
		fatal("Disagreement in the number of contigs");
}

#define BIG_BUF_SIZE 1000000
#define MED_BUF_SIZE 10000
#define LITTLE_BUF_SIZE 1000

Block_ptr read_blocks(char *filename, Seq_ptr *seq_ptr_ptr, int *Nseq,
  Contig_ptr *contig_ptr_ptr, int *Ncontig)
{
	FILE *fp;
	char *buf, little_buf[LITTLE_BUF_SIZE], *lastline, *p, *q, sign;
	uchar *t, *f;
	int nseq, ncontig, i, j, k, bufsize, verbose;
	Seq_ptr seq;
	Contig_ptr contig;
	Seg_ptr sp;
	Block_ptr blocks = NULL, b, last_b = NULL;

	fp = ckopen(filename, "r");
	if (fgets(little_buf, LITTLE_BUF_SIZE, fp) == NULL ||
	    strncmp(little_buf, "#:bb",4))
		fatal("Not a blockbuilder output file.");
	verbose = (strstr(little_buf, "verbose") != NULL);
	if (fgets(little_buf, LITTLE_BUF_SIZE, fp) == NULL ||
	    strncmp(little_buf, "blockbuilder", 12) != 0)
		fatal("Not a proper blockbuilder output file.");
	bufsize = (verbose ? BIG_BUF_SIZE : MED_BUF_SIZE);
	
	buf = ckalloc (bufsize * sizeof (char));
	fgets(buf, bufsize, fp);
	if (fgets(buf, bufsize, fp) == NULL ||
	    sscanf(buf, "%d %d", &nseq, &ncontig) != 2)
		fatalf("cannot get number of sequences and contigs: %s", buf);
	*Nseq = nseq;
	*Ncontig = ncontig;
	*seq_ptr_ptr = seq = ckalloc(nseq * sizeof(Seq));
	*contig_ptr_ptr = contig = ckalloc(ncontig * sizeof(Contig));
	for (i = -1, j = 0; j < ncontig; ++j) {
		if (fgets(buf, bufsize, fp) == NULL)
			fatal("read_blocks: cannot happen");
		p = buf;
		while (*p && *p != ' ')
			++p;
		if (*p != ' ')
			fatal("Cannot find blank");
		*p = '\0';
		if (i == -1 || !same_string(buf, seq[i].name))
			seq[++i].name = copy_string(buf);
		if (*++p != '"')
			fatal("Cannot find first quote.");
		if ((q = strchr(++p, '"')) == NULL)
			fatal("Cannot find second quote.");
		*q = '\0';
		contig[j].seq_nbr = i;
		contig[j].header = copy_string(p);
		contig[j].len = atoi(q+1);
	}

	read_seqs(seq, nseq, contig, ncontig);
	if ((lastline = fgets(buf, bufsize, fp)) == NULL)
		fatal("no blocks were given");
	while (lastline) {
		b = ckalloc(sizeof(Block));
		b->next = NULL;
		if (blocks == NULL)
			blocks = b;
		else
			last_b->next = b;
		last_b = b;
		if (fgets(buf, bufsize, fp) == NULL ||
		   sscanf(buf, "%d %d", &(b->nrow), &(b->ncol)) != 2)
			fatalf("nrow, ncol: %s", buf);
		b->seg = ckalloc(b->nrow * sizeof(Seg));
		i = -1;
		while ((lastline = fgets(buf, bufsize, fp)) != NULL &&
		    strcmp(buf, "\n")) {
			sp = &(b->seg[++i]);
			if (sscanf(buf, "%d %c %d-%d", &(sp->contig_nbr),
			   &sign, &(sp->beg), &(sp->end)) != 4)
				fatalf("contig_nbr, sign, beg, end: %s", buf);
			t = sp->chars = ckalloc((b->ncol + 1) * sizeof(uchar));
			q = buf;
			while (isdigit((uchar)*q))
				++q;
			while (!isdigit((uchar)*q))
				++q;
			p = strchr(q, '-') + 1;
			while (isdigit((uchar)*p))
				++p;
			++p;
			if (verbose) {
				for (k = 0; k < b->ncol; ++k)
					*t++ = *p++;
			} else {
				f = contig[sp->contig_nbr].chars;
				f += sp->beg - 1;
				while ((k = atoi(p)) != 0) {
					if (k < 0)
						while (k++ < 0)
							*t++ = '-';
					else
						while (k-- > 0)
							*t++ = *f++;
					while (isspace((uchar)*p))
						++p;
					while (*p == '-' || isdigit((uchar)*p))
						++p;
				}
			}
			if (t - sp->chars != b->ncol) {
				fprintf(stderr,
				   "t - sp->chars = %d, b->ncol = %d\n",
			    	t - sp->chars, b->ncol);
				fatalf("read_blocks: incorrect seg length: %s",
					buf);
			}
			sp->chars[b->ncol] = '\0';
		}
		if (i != b->nrow - 1)
			fatal("read_blocks: problem with nrow");
	}

	free(buf);
	
	return blocks;
}

Block_ptr block_excise(Block_ptr b, Block_ptr blocks)
{
	Block_ptr p;

	if (blocks == NULL)
		fatal("block_delete: empty list of blocks");
	if (blocks == b)
		blocks = b->next;
	else {
		for (p = blocks; (p->next) != NULL; p = p->next)
			if (p->next == b)
				break;
		if (p->next == NULL)
			fatal("block_delete cannot find block");
		p->next = b->next;
	}
	return blocks;
}

void block_free(Block_ptr b)
{
	int i;

	for (i = 0; i < b->nrow; ++i)
		free (b->seg[i].chars);
	free(b->seg);
	free(b);
}

/* reduce-block - eliminate trivial rows (i.e., all dashes) from a block */
void reduce_block(Block_ptr b, Contig_ptr contig)
{
	int i, j;
	Seg_ptr s;
	uchar *t, *u;

	/* i is current row, j is where the next non-trivial row goes */
	for (i = j = 0; i < b->nrow; ++i) {
		s = &(b->seg[i]);
		for (t = u = s->chars; t - u < b->ncol; ++t)
			if (*t != '-')
				break;
		if (t - u == b->ncol) {
			if (s->end != s->beg - 1)
				fatal("trivial row has non-zero length");
			fprintf(stderr,
			    "trivial row of width %d in `%s' at position %d\n",
			     b->ncol, contig[b->seg[i].contig_nbr].header,
			     s->beg);
				free(b->seg[i].chars);
		} else {
			if (j < i)
				b->seg[j] = b->seg[i];
			++j;
		}
	}
	if (j == 0)
		fatal("reduce-block: no rows remain in block");
	b->nrow = j;
}

/* get_row_from_seq_num -- given the sequence number, determine the row in the 
	block of the segment */
int get_row_from_seq_num(int seq_num, int nrow, Block_ptr block)
{
    	int j, row = -1;

    	for (j = 0; j < nrow; j++)
        	if (block->seg[j].contig_nbr == seq_num)
        	{
            	   	row = j;
            		break;
        	}

    	return row;
}
